package controller;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.util.LinkedList;
import javax.swing.DefaultComboBoxModel;
import javax.swing.JComboBox;
import javax.swing.event.MouseInputListener;
import controller.listener.AntiAliasedLineListener;
import controller.listener.CircleListener;
import controller.listener.ClippingRectangleListener;
import controller.listener.LineListener;
import controller.listener.PointListener;
import controller.listener.PolygonListener;
import controller.listener.RectangleListener;
import model.drawables.DrawableObject;
import model.drawables.Polygon;
import model.drawables.Rectangle;
import view.DrawingPanelView;

/**
 * Der Controller, der eine DrawingPanelView verwaltet und das dazugehörige
 * Model in Form einer Liste von DrawableObjects. Das Model des Controllers ist
 * hart mit dem der View verbunden. Änderungen werden also direkt in der View
 * passieren. Ggf. muss die View aufgefordert werden sich neu zu zeichnen.
 * 
 * @author Nicolas Neubauer
 * 
 */
public class DrawingPanelViewController implements DrawableObjectProcessing {

	// View
	private DrawingPanelView drawingPanelView;

	// Model
	private LinkedList<DrawableObject> drawableObjectsModel;
	private DrawableObject temporaryObject;

	// Listener für die einzelnen grafischen Objekte
	private PointListener pointListener;
	private LineListener lineListener;
	private AntiAliasedLineListener antiAliasedLineListener;
	private RectangleListener rectangleListener;
	private CircleListener circleListener;
	private PolygonListener polygonListener;
	private PolygonListener filledPolygonListener;
	private ClippingRectangleListener clippingRectangleListener;

	/**
	 * Der Konstruktor initialisiert die View und legt die Listener an.
	 */
	public DrawingPanelViewController() {

		drawableObjectsModel = new LinkedList<DrawableObject>();
		drawingPanelView = new DrawingPanelView(640, 480, drawableObjectsModel);

		// das UI anpassen
		drawingPanelView.getButton().setText("Clear");

		// welche Objekte sollen gezeichnet werden können?
		String[] objectArray = { "Punkt", "Linie", "Anti-Aliased Linie",
				"Rechteck", "Polygon", "gefuelltes Polygon", "Kreis",
				"Polygon-Clipping" };
		drawingPanelView.getComboBox().setModel(
				new DefaultComboBoxModel(objectArray));

		// Listener für die einzelnen Grafischen Objekte anlegen.
		pointListener = new PointListener(this);
		lineListener = new LineListener(this);
		antiAliasedLineListener = new AntiAliasedLineListener(this);
		rectangleListener = new RectangleListener(this);
		circleListener = new CircleListener(this);
		polygonListener = new PolygonListener(this, false);
		filledPolygonListener = new PolygonListener(this, true);
		clippingRectangleListener = new ClippingRectangleListener(this);

		// Event-Listener fuer Button
		ActionListener a = new ActionListener() {

			public void actionPerformed(ActionEvent e) {
				// Das Model leeren und die View neu zeichnen
				drawableObjectsModel.clear();
				drawingPanelView.getDrawingPanel().repaint();
			}

		};

		// ChangeListener um festzustellen, was gezeichnet werden soll
		ActionListener c = new ActionListener() {

			/**
			 * Hängt den passenden Listener an die Zeichenfläche
			 */
			public void actionPerformed(ActionEvent e) {
				// Hänge das aktuell gewählte Objekt als Listener an die
				// View

				JComboBox cb = (JComboBox) e.getSource();

				if (cb.getSelectedItem().equals("Punkt"))
					changeMouseInputListenerTo(pointListener);
				else if (cb.getSelectedItem().equals("Linie"))
					changeMouseInputListenerTo(lineListener);
				else if (cb.getSelectedItem().equals("Anti-Aliased Linie"))
					changeMouseInputListenerTo(antiAliasedLineListener);
				else if (cb.getSelectedItem().equals("Rechteck"))
					changeMouseInputListenerTo(rectangleListener);
				else if (cb.getSelectedItem().equals("Kreis"))
					changeMouseInputListenerTo(circleListener);
				else if (cb.getSelectedItem().equals("Polygon"))
					changeMouseInputListenerTo(polygonListener);
				else if (cb.getSelectedItem().equals("gefuelltes Polygon"))
					changeMouseInputListenerTo(filledPolygonListener);
				else if (cb.getSelectedItem().equals("Polygon-Clipping"))
					changeMouseInputListenerTo(clippingRectangleListener);
			}

			/**
			 * Convenience-Methode zum umsetzen des aktuellen MouseListeners an
			 * der Zeichenfläche der verwalteten View
			 * 
			 * @param newListener
			 *            der zu setzende Listener, alle anderen werden
			 *            ausgehängt
			 */
			private void changeMouseInputListenerTo(
					MouseInputListener newListener) {

				// InputListeners
				MouseListener[] currentInputListeners = drawingPanelView
						.getDrawingPanel().getMouseListeners();

				for (MouseListener curListener : currentInputListeners)
					drawingPanelView.getDrawingPanel().removeMouseListener(
							curListener);

				// MotionListeners
				MouseMotionListener[] currentMotionListners = drawingPanelView
						.getDrawingPanel().getMouseMotionListeners();
				for (MouseMotionListener curListener : currentMotionListners)
					drawingPanelView.getDrawingPanel()
							.removeMouseMotionListener(curListener);

				drawingPanelView.getDrawingPanel()
						.addMouseListener(newListener);
				drawingPanelView.getDrawingPanel().addMouseMotionListener(
						newListener);
			}

		};

		// Anhängen der Listener
		drawingPanelView.getButton().addActionListener(a);
		drawingPanelView.getComboBox().addActionListener(c);

		// Event einmal auslösen, damit der korrekte Listener angehängt wird
		drawingPanelView.getComboBox().setSelectedItem("Punkt");

	}

	/**
	 * Gibt die verwaltete View zurück
	 * 
	 * @return die View
	 */
	public DrawingPanelView getView() {
		return drawingPanelView;
	}

	/**
	 * Fügt das DrawableObject in das Model ein und fordert die View zum
	 * Neuzeichnen auf.
	 * 
	 * @param drawableObject
	 */
	public void processDrawableObject(DrawableObject drawableObject) {

		drawableObjectsModel.add(drawableObject);
		drawingPanelView.getDrawingPanel().repaint();

	}

	/**
	 * Löscht das temporäre DrawableObject aus dem Model und fordert die View
	 * zum Neuzeichnen auf.
	 */
	public void clearTemporaryDrawableObject() {

		drawableObjectsModel.remove(temporaryObject);
		drawingPanelView.getDrawingPanel().repaint();
		temporaryObject = null;

	}

	/**
	 * Fügt ein temporäres Objekt in das Model ein und fordert die View zum
	 * Neuzeichnen auf. Das Objekt kann mit clearTemporaryDrawableObject wieder
	 * aus dem Model entfernt werden.
	 * 
	 * @param drawableObject
	 */
	public void setTemporaryDrawableObject(DrawableObject drawableObject) {

		// Erst löschen
		if (temporaryObject != null)
			this.clearTemporaryDrawableObject();

		temporaryObject = drawableObject;
		this.processDrawableObject(temporaryObject);

	}

	/**
	 * Clippt alle im Model verwalteten Polygone mit deren clip()-Methode und
	 * entfernt ggf. nicht mehr sinnvoll definierte Polygone aus dem Model.
	 * Initiiert außerdem das Neu-Zeichnen der Zeichenfläche
	 * 
	 * @param r
	 */
	public void clipDrawableObjects(Rectangle r) {
		boolean stillPolygon;
		LinkedList<DrawableObject> toBeDeleted = new LinkedList<DrawableObject>();

		if (!drawableObjectsModel.isEmpty()) {
			for (DrawableObject d : drawableObjectsModel) {
				if (d instanceof Polygon) {
					Polygon p = (Polygon) d;
					stillPolygon = p.clip(r);
					if (!stillPolygon) {
						toBeDeleted.add(p);
					}
				}
			}
		}

		drawableObjectsModel.removeAll(toBeDeleted);
		drawingPanelView.getDrawingPanel().repaint();
	}
}
