/**
 * PolygonListener
 * @author Denis Meyer
 * @author Niels Meyering
 */
package controller.listener;

import java.awt.event.MouseEvent;
import javax.swing.event.MouseInputAdapter;
import controller.DrawableObjectProcessing;
import model.drawables.DashedPolygon;
import model.drawables.Line;
import model.drawables.Point;
import model.drawables.Polygon;

/**
 * Ein Listener, der Lines durch MouseEvents erzeugt und an ein Objekt, das
 * DrawableObjectProcessing implementiert übergibt.
 * 
 * @author Nicolas Neubauer
 * @author Denis Meyer
 * @author Niels Meyering
 * 
 */
public class PolygonListener extends MouseInputAdapter {

	private DrawableObjectProcessing delegate;

	// Startpunktsumgebung, damit Polygon sich schliesst
	private int epsilon = 4;
	private Polygon polygon;
	private DashedPolygon dashedPolygon;
	private boolean filled;

	/**
	 * Konstruktor
	 * 
	 * @param delegate
	 * @param filled
	 */
	public PolygonListener(DrawableObjectProcessing delegate, boolean filled) {
		this.delegate = delegate;
		this.filled = filled;
	}

	/**
	 * 
	 * @return Delegate
	 */
	public DrawableObjectProcessing getDelegate() {
		return delegate;
	}

	/**
	 * Wenn Benutzer Maus geklickt hat
	 * 
	 * @param e
	 */
	public void mouseClicked(MouseEvent e) {
		Point p = new Point(e.getX(), e.getY()); // Startpunkt

		if (polygon == null) {
			polygon = new Polygon(p, filled); // Adde Punkt P zu Polygonpunkten
			dashedPolygon = new DashedPolygon(p, false);
		} else {
			// Wenn Startpunkt erreicht und Polygon hat mehr als 2 Punkte
			if (nearStart(p) && (polygon.getNumberOfPoints() > 2)) {
				// Setze letzte Linie bis zum Startpunkt
				delegate.clearTemporaryDrawableObject();
				polygon.close(); // Polygon schliessen
				delegate.processDrawableObject(polygon);
				// Alles zurcksetzen
				polygon = null;
			} else if (new Line(polygon.getPointAt(0), new Point(e.getX(), e
					.getY())).length() > epsilon) {
				// nchste Polygonlinie zeichnen
				polygon.addToList(p); // Fge Punkt zur Polygonliste
				dashedPolygon.addToList(p);
				// delegate.processDrawableObject(dashedPolygon);
				delegate.setTemporaryDrawableObject(dashedPolygon);
			}
		}
	}

	/**
	 * Gibt zurck, ob der Punkt end nahe genug am Startpunkt liegt, damit das
	 * Polygon als geschlossen betrachtet werden kann
	 * 
	 * @param end
	 * @return ob end nahe genug am Startpunkt liegt, damit das Polygon als
	 *         geschlossen betrachtet werden kann
	 */
	public boolean nearStart(Point end) {
		// Wenn Polygon mehr als 2 Punkte besitzt
		if (polygon.getNumberOfPoints() > 2) {
			// Hole den Startpunkt
			Point start = polygon.getPointAt(0);
			// Schaue, ob gerade gesetzter Punkt in Umgebung von Startpunkt
			return (Math.abs(start.x - end.x) <= epsilon)
					&& (Math.abs(start.y - end.y) <= epsilon);
		}
		return false;
	}

	/**
	 * Maus hat sich bewegt
	 * 
	 * @param e
	 */
	public void mouseMoved(MouseEvent e) {
		// Wenn ein Polygon vorhanden ist
		if (polygon != null) {
			// Erstelle neues temporres Polygon
			DashedPolygon dpt = new DashedPolygon(dashedPolygon);
			// Fge aktuelle Mauskoordinate hinzu
			dpt.addToList(new Point(e.getX(), e.getY()));
			// Zeichne gestricheltes Polygon
			delegate.setTemporaryDrawableObject(dpt);
		}
	}
}
