package model.drawables;

import java.awt.Graphics;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;

/**
 * Klasse, die ein Polygon repr&auml;sentiert.
 * 
 * @author Denis Meyer
 * @author Niels Meyering
 */
public class Polygon extends DrawableObject {
	// ArrayList, um die Polygonpunkte zu speichern
	private ArrayList<Point> punkte;
	// Flag, ob das Polygon geschlossen ist oder nicht
	private boolean geschlossen;
	// Flag, ob das Polygon gefllt werden soll oder nicht
	private boolean gefuellt;

	/**
	 * Konstruktor zum Erzeugen eines Polygons
	 * 
	 * @param start
	 *            Startpunkt
	 * @param filled
	 *            soll gefuellt werden oder nicht
	 */
	public Polygon(Point start, boolean filled) {
		punkte = new ArrayList<Point>();
		geschlossen = false; // Polygon zu Anfang nicht geschlossen
		addToList(start); // Adde start zu Polygonpunkten
		gefuellt = filled;
	}

	/**
	 * Copy-Konstructor
	 * 
	 * @param p
	 *            Copy-Konstructor
	 */
	public Polygon(Polygon p) {
		punkte = new ArrayList<Point>();
		for (Point pg : p.getPoints()) {
			addToList(pg);
		}
		geschlossen = p.geschlossen;
		gefuellt = p.gefuellt;
	}

	/**
	 * Gibt Polygonpunkte zurck
	 * 
	 * @return Polygonpunkte
	 */
	public ArrayList<Point> getPoints() {
		return punkte;
	}

	/**
	 * Gibt an, wie viele Punkte die Polygonliste hat
	 * 
	 * @return wie viele Punkte die Polygonliste hat
	 */
	public int getNumberOfPoints() {
		if (getPoints() != null) {
			return getPoints().size();
		}
		return 0;
	}

	/**
	 * Gibt Punkt an Stelle n zurck
	 * 
	 * @param n
	 *            Stelle n
	 * @return Punkt an Stelle n oder null
	 */
	public Point getPointAt(int n) {
		// Wenn n im richtigen Bereich
		if ((getPoints() != null) && (n >= 0) && (n < getNumberOfPoints())) {
			return getPoints().get(n);
		}
		return null;
	}

	/**
	 * Schliesst das Polygon: Verbindet 1. mit letztem Punkt der Liste
	 */
	public void close() {
		// Wenn Punktarray vorhanden
		if (getPoints() != null) {
			// Wenn Polygon schon geschlossen und erster und letzter Punkt
			// bereinstimmen
			if (closed()
					&& (getPointAt(0) == getPointAt(getNumberOfPoints() - 1))) {
				return;
				// Sonst, wenn mindestens 3 Punkte vorhanden sind
			} else if (getNumberOfPoints() > 2) {
				// Polygon schlieen
				addToList(new Point(getPointAt(0)));
				geschlossen = true;
			}
		}
	}

	/**
	 * Gibt zurck, ob das Polygon geschlossen ist oder nicht
	 * 
	 * @return ob das Polygon geschlossen ist oder nicht
	 */
	public boolean closed() {
		return geschlossen;
	}

	/**
	 * Gibt zurueck, ob Polygon gefllt werden soll oder nicht
	 * 
	 * @return ob Polygon gefllt werden soll oder nicht
	 */
	public boolean filled() {
		return gefuellt;
	}

	/**
	 * Paint-Methode
	 * 
	 * @param g
	 *            der Graphikkontext, in den das Objekt gezeichnet werden soll
	 */
	public void paint(Graphics g) {
		Line l;
		Point prev = null;

		// Verbindet einen Punkt mit dem Darauffolgenden
		for (Point p : getPoints()) {
			if (prev != null) {
				l = new Line(prev, p);
				l.paint(g);
			}
			prev = p;
		}

		if (filled() && closed()) {
			scanLineFill(g);
		}
	}

	/**
	 * Packt Punkt p mit in die Polygonliste
	 * 
	 * @param p
	 */
	public void addToList(Point p) {
		if (getPoints() != null) {
			if (closed() && !getPoints().isEmpty()) {
				// Entfernt letzten Punkt, wenn geschlossen
				removeFromList(getPoints().size() - 1);
			}
			// Added Punkt zu Polygonpunktenliste
			getPoints().add(p);
		}
	}

	/**
	 * Lscht Index i aus der Polygonliste
	 * 
	 * @param i
	 */
	public void removeFromList(int i) {
		if (getPoints() != null) {
			if ((i >= 0) && (i < getNumberOfPoints())) {
				getPoints().remove(i);
			}
		}
	}

	/**
	 * Wendet das Scanline-Verfahren zum Fuellen eines Polygons an
	 * 
	 * @param g
	 *            der grafische Kontext, in den dieses Objekt sich zeichnen soll
	 */
	private void scanLineFill(Graphics g) {
		// nach y sort. Kantenliste
		ArrayList<Line> sortierteKanten = new ArrayList<Line>();
		// Liste der aktiven Kanten
		ArrayList<Line> aktiveKanten = new ArrayList<Line>();
		// Liste der Schnittpunkte
		ArrayList<Point> schnittpunkte = new ArrayList<Point>();
		boolean leer = false; // akt. Kantenliste noch leer

		// Starte die ScanLine
		// - von Punkt 1 aus ber y-Koordinate laufen
		// - alle aktiven Kanten abarbeiten
		sortierteKanten = getSortedEdges();
		if (!sortierteKanten.isEmpty()) {
			for (int line = sortierteKanten.get(0).a.y; !leer; line--) {
				// Hole aktuell aktive Kantenliste
				aktiveKanten = getActivePolEdges(sortierteKanten, aktiveKanten,
						line);
				// Berechne Schnittpunkte der aktuellen Kanten mit der Scanline
				schnittpunkte = getSortedIntersectionsWithSL(aktiveKanten, line);
				// Zeichne im Polygon liegende Scanline
				paintPolSLIntersections(schnittpunkte, g);
				// berprfe, ob noch aktive Kanten vorhanden sind
				leer = aktiveKanten.isEmpty();
			}
		}
	}

	/**
	 * Gibt den horizontalen/vertikalen Schnittpunkt zurck
	 * 
	 * @param l1
	 * @param yx
	 * @return den horizontalen/vertikalen Schnittpunkt, wenn l != null, null
	 *         sonst
	 */
	private Point getIntersection(Line l, int yx, boolean horiz) {
		if (l != null) {
			if (horiz && (l.a.x == l.e.x)) {
				return new Point(l.a.x, yx);
			} else if (!horiz && (l.a.y == l.e.y)) {
				return new Point(yx, l.a.y);
			}
			// Steigung der Geraden
			double m = ((double) l.a.y - l.e.y) / (l.a.x - l.e.x);
			// Geradengleichung: f(x) = y = m * x + b
			// b = y - (m * x)
			double b = l.a.y - (m * l.a.x);
			if (horiz) {
				// x = (y - b) / m
				return new Point((int) ((yx - b) / m), yx);
			}
			return new Point(yx, (int) (m * yx + b));
		}
		return null;
	}

	/**
	 * Liefert Alle Kanten des Polygons sortiert in einer Liste. - y-Wert von
	 * Anfangspkt. > y-Wert Endpunkt (absteigend sortiert) - keine waagerechten
	 * Kanten - grter y-Wert zuerst
	 * 
	 * @return Kantenliste
	 */
	private ArrayList<Line> getSortedEdges() {
		Point prev = null;
		ArrayList<Line> kanten = new ArrayList<Line>();

		if (getPoints() != null) {
			for (Point p : getPoints()) {
				// Keine waagerechten Linien
				if ((prev != null) && (prev.y != p.y)) {
					// Reihenfolge der y-Werte des Anfangs- und Endpunktes
					// beachten
					// und jeweils Linie erstellen
					if (prev.y > p.y) {
						kanten.add(new Line(prev, p));
					} else {
						kanten.add(new Line(p, prev));
					}
				}
				prev = p;
			}

			Collections.sort(kanten); // Liste sortieren
		}
		return kanten;
	}

	/**
	 * Gibt aktuelle Liste mit aktiven Kanten zurck. - Kanten der vorherigen
	 * ScanLine verschwinden - Neue Kanten werden hinzugefgt
	 * 
	 * @param edges
	 *            Polygonkanten
	 * @param activeEdges
	 *            aktive Kanten
	 * @param line
	 *            Scanline
	 * @return aktuelle Liste mit aktiven Kanten
	 */
	private ArrayList<Line> getActivePolEdges(ArrayList<Line> kanten,
			ArrayList<Line> aktiveKanten, int scanline) {
		if (kanten != null) {
			for (Line l : kanten) {
				// Linie entfernen, wenn Ende der Linie auf ScanLine gefallen
				// ist
				if (l.e.y == scanline) {
					aktiveKanten.remove(l);
				}
				// Linie hinzufgen, wenn Anfang der Linie auf ScanLine gefallen
				// ist
				if (l.a.y == scanline) {
					aktiveKanten.add(l);
				}
			}
		}
		return aktiveKanten;
	}

	/**
	 * Liefert Liste mit sortierten Schnittpunkten der aktuellen Kantenliste
	 * zurck
	 * 
	 * @param activeEdges
	 *            aktive Kanten
	 * @param line
	 *            Scanline
	 * @return Liste mit sortierten Schnittpunkten der aktuellen Kantenliste
	 */
	private ArrayList<Point> getSortedIntersectionsWithSL(
			ArrayList<Line> aktiveKanten, int scanline) {
		ArrayList<Point> schnittpunkte = new ArrayList<Point>();

		if (aktiveKanten != null) {
			for (Line l : aktiveKanten) {
				// Fge neuen Punkt hinzu
				schnittpunkte.add(getIntersection(l, scanline, true));
			}

			// Schnittpunkte nach x-Koordinate sortieren
			Collections.sort(schnittpunkte);
		}
		return schnittpunkte;
	}

	/**
	 * Durchlaufe Schnittpunktlinie und zeichne fuer alle Paare von Punkten eine
	 * Linie.
	 * 
	 * @param intersections
	 *            Schnittpunktmenge
	 * @param g
	 *            der grafische Kontext, in den die Linien gezeichnet werden
	 *            sollen
	 */
	private void paintPolSLIntersections(List<Point> schnittpunkte, Graphics g) {
		if (schnittpunkte != null) {
			Iterator<Point> it = schnittpunkte.iterator();
			Point start;
			while (it.hasNext()) {
				// Setze Startpunkt
				start = it.next();
				// Wenn noch ein Punkt vorhanden ist
				if (it.hasNext()) {
					// Setze Endpunkt und zeichne Linie
					new Line(start, it.next()).paint(g);
				}
			}
		}
	}

	/**
	 * Clipt das Polygon im Rechteck r
	 * 
	 * @param r
	 * @return true, wenn das geclipte noch ein Polygon ist, false sonst
	 */
	public boolean clip(Rectangle r) {
		if (r != null) {
			// Clip linke Seite
			clipLeft(r.a.x);
			// Clip obere Seite
			clipTop(r.a.y);
			// Clip rechte Seite
			clipRight(r.e.x);
			// Clip untere Seite
			clipDown(r.e.y);
			if (getNumberOfPoints() > 2) {
				return true;
			} else {
				return false;
			}
		}
		return true;
	}

	/**
	 * Clipt links
	 * 
	 * @param line
	 *            linke x-Koordinate
	 */
	private void clipLeft(int line) {
		Point letzterPunkt = null;
		// Liste der neuen Polygonpunkte
		ArrayList<Point> neuePunkte = new ArrayList<Point>();
		for (Point aktuellerPunkt : getPoints()) {
			// Anfangspunkt setzen
			if (letzterPunkt == null) {
				letzterPunkt = aktuellerPunkt;
			} else {
				// Wenn letzter Punkt im sichtbaren Bereich
				if (letzterPunkt.x >= line) {
					// Fge ihn in die Liste der neuen Polygonpunkte ein
					neuePunkte.add(letzterPunkt);
					// Wenn aktueller Punkt nicht im sichtbaren Bereich
					if (aktuellerPunkt.x < line) {
						// Fge den Schnittpunkt vom letzten zum aktuellen Punkt
						// in die Liste der neuen Polygonpunkte ein
						neuePunkte.add(getIntersection(new Line(letzterPunkt,
								aktuellerPunkt), line, false));
					}
					// Wenn letzter Punkt nicht im sichtbaren Bereich und
					// aktueller Punkt im sichtbaren Bereich
				} else if (aktuellerPunkt.x >= line) {
					// Fge den Schnittpunkt vom letzten zum aktuellen Punkt
					// in die Liste der neuen Polygonpunkte ein
					neuePunkte.add(getIntersection(new Line(letzterPunkt,
							aktuellerPunkt), line, false));
				}
			}
			// Setze aktuellen Punkt als neuen letzten Punkt
			letzterPunkt = aktuellerPunkt;
		}

		// Setze Punkte fr neues Polygon
		geschlossen = false;
		punkte = neuePunkte;
		close();
	}

	/**
	 * Clipt (vom Nutzer aus gesehen) oben
	 * 
	 * @param line
	 *            obere y-Koordinate
	 */
	private void clipTop(int line) {
		Point letzterPunkt = null;
		// Liste der neuen Polygonpunkte
		ArrayList<Point> neuePunkte = new ArrayList<Point>();
		for (Point aktuellerPunkt : getPoints()) {
			// Anfangspunkt setzen
			if (letzterPunkt == null) {
				letzterPunkt = aktuellerPunkt;
			} else {
				// Wenn letzter Punkt im sichtbaren Bereich
				if (letzterPunkt.y >= line) {
					// Fge ihn in die Liste der neuen Polygonpunkte ein
					neuePunkte.add(letzterPunkt);
					// Wenn aktueller Punkt nicht im sichtbaren Bereich
					if (aktuellerPunkt.y < line) {
						// Fge den Schnittpunkt vom letzten zum aktuellen Punkt
						// in die Liste der neuen Polygonpunkte ein
						neuePunkte.add(getIntersection(new Line(letzterPunkt,
								aktuellerPunkt), line, true));
					}
					// Wenn letzter Punkt nicht im sichtbaren Bereich und
					// aktueller Punkt im sichtbaren Bereich
				} else if (aktuellerPunkt.y >= line) {
					// Fge den Schnittpunkt vom letzten zum aktuellen Punkt
					// in die Liste der neuen Polygonpunkte ein
					neuePunkte.add(getIntersection(new Line(letzterPunkt,
							aktuellerPunkt), line, true));
				}
			}
			// Setze aktuellen Punkt als neuen letzten Punkt
			letzterPunkt = aktuellerPunkt;
		}

		// Setze Punkte fr neues Polygon
		geschlossen = false;
		punkte = neuePunkte;
		close();
	}

	/**
	 * Clipt rechts
	 * 
	 * @param line
	 *            obere x-Koordinate
	 */
	private void clipRight(int line) {
		Point letzterPunkt = null;
		// Liste der neuen Polygonpunkte
		ArrayList<Point> neuePunkte = new ArrayList<Point>();
		for (Point aktuellerPunkt : getPoints()) {
			// Anfangspunkt setzen
			if (letzterPunkt == null) {
				letzterPunkt = aktuellerPunkt;
			} else {
				// Wenn letzter Punkt im sichtbaren Bereich
				if (letzterPunkt.x <= line) {
					// Fge ihn in die Liste der neuen Polygonpunkte ein
					neuePunkte.add(letzterPunkt);
					// Wenn aktueller Punkt nicht im sichtbaren Bereich
					if (aktuellerPunkt.x > line) {
						// Fge den Schnittpunkt vom letzten zum aktuellen Punkt
						// in die Liste der neuen Polygonpunkte ein
						neuePunkte.add(getIntersection(new Line(letzterPunkt,
								aktuellerPunkt), line, false));
					}
					// Wenn letzter Punkt nicht im sichtbaren Bereich und
					// aktueller Punkt im sichtbaren Bereich
				} else if (aktuellerPunkt.x <= line) {
					// Fge den Schnittpunkt vom letzten zum aktuellen Punkt
					// in die Liste der neuen Polygonpunkte ein
					neuePunkte.add(getIntersection(new Line(letzterPunkt,
							aktuellerPunkt), line, false));
				}
			}
			// Setze aktuellen Punkt als neuen letzten Punkt
			letzterPunkt = aktuellerPunkt;
		}

		// Setze Punkte fr neues Polygon
		geschlossen = false;
		punkte = neuePunkte;
		close();
	}

	/**
	 * Clipt (vom Nutzer aus gesehen) unten
	 * 
	 * @param line
	 *            untere y-Koordinate
	 */
	private void clipDown(int line) {
		Point letzterPunkt = null;
		// Liste der neuen Polygonpunkte
		ArrayList<Point> neuePunkte = new ArrayList<Point>();
		for (Point aktuellerPunkt : getPoints()) {
			// Anfangspunkt setzen
			if (letzterPunkt == null) {
				letzterPunkt = aktuellerPunkt;
			} else {
				// Wenn letzter Punkt im sichtbaren Bereich
				if (letzterPunkt.y <= line) {
					// Fge ihn in die Liste der neuen Polygonpunkte ein
					neuePunkte.add(letzterPunkt);
					// Wenn aktueller Punkt nicht im sichtbaren Bereich
					if (aktuellerPunkt.y > line) {
						// Fge den Schnittpunkt vom letzten zum aktuellen Punkt
						// in die Liste der neuen Polygonpunkte ein
						neuePunkte.add(getIntersection(new Line(letzterPunkt,
								aktuellerPunkt), line, true));
					}
					// Wenn letzter Punkt nicht im sichtbaren Bereich und
					// aktueller Punkt im sichtbaren Bereich
				} else if (aktuellerPunkt.y <= line) {
					// Fge den Schnittpunkt vom letzten zum aktuellen Punkt
					// in die Liste der neuen Polygonpunkte ein
					neuePunkte.add(getIntersection(new Line(letzterPunkt,
							aktuellerPunkt), line, true));
				}
			}
			// Setze aktuellen Punkt als neuen letzten Punkt
			letzterPunkt = aktuellerPunkt;
		}

		// Setze Punkte fr neues Polygon
		geschlossen = false;
		punkte = neuePunkte;
		close();
	}
}
